@@ -8,6 +8,7 @@ use std::io::Write;
 use regex::Regex;
 use ansi_term::Color;
 use ansi_term::Color::Fixed;
+use unicode_segmentation::UnicodeSegmentation;
 
 
 macro_rules! skip_bad_range {
@@ -69,7 +70,9 @@ impl Byte {
         use crate::ByteCategory::*;
 
         match self.category() {
-            Null => '0',
+
+            /* hexyl uses 0 here, depending on color to distinguish */
+            Null => '•',
             AsciiPrintable => self.0 as char,
             AsciiWhitespace if self.0 == 0x20 => ' ',
             AsciiWhitespace => '_',
@@ -105,6 +108,7 @@ i          Prompt you to write out bytes which will be inserted at current index
              inserted there.
 12,3dp     Print bytes 12 - 3d inclusive, move to leftmost byte printed on the
              last line.
+m          Toggle whether or not characters are printed after bytes
 n          Toggle whether or not byte numbers are printed before bytes
 p          Print current line of byte(s) (depending on 'W')
 s          Print state of all toggles and 'W'idth
@@ -157,7 +161,7 @@ impl Command {
         let re_blank_line = Regex::new(r"^ *$").unwrap();
         let re_plus = Regex::new(r"^ *\+ *$").unwrap();
         let re_minus = Regex::new(r"^ *\- *$").unwrap();
-        let re_single_char_command = Regex::new(r"^ *(?P<command>[?npsxqwik]).*$").unwrap();
+        let re_single_char_command = Regex::new(r"^ *(?P<command>[?mnpsxqwik]).*$").unwrap();
         let re_range = Regex::new(r"^ *(?P<begin>[0-9a-fA-F.$]+) *, *(?P<end>[0-9a-fA-F.$]+) *(?P<the_rest>.*) *$").unwrap();
         let re_specified_index = Regex::new(r"^ *(?P<index>[0-9A-Fa-f.$]+) *(?P<the_rest>.*) *$").unwrap();
         let re_offset_index = Regex::new(r"^ *(?P<sign>[-+])(?P<offset>[0-9A-Fa-f]+) *(?P<the_rest>.*) *$").unwrap();
@@ -444,6 +448,16 @@ fn formatted_byte(byte:u8, color:bool) -> String {
 }
 
 
+fn chared_byte(byte:u8, color:bool) -> String {
+    if color {
+        Byte(byte).color().paint(String::from(Byte(byte).as_char())).to_string()
+    }
+    else {
+        Byte(byte).as_char().to_string()
+    }
+}
+
+
 fn padded_byte(byte:u8) -> String {
     format!("{:02x}", byte)
 }
@@ -459,19 +473,33 @@ fn max_bytes_line(bytes:&[u8], width:NonZeroUsize) -> usize {
 }
 
 
-fn bytes_line(bytes:&[u8], line_number:usize, width:NonZeroUsize) -> &[u8] {
+fn bytes_line_range(bytes:&[u8], line_number:usize, width:NonZeroUsize) -> std::ops::Range<usize> {
     let width = usize::from(width);
     if line_number * width < bytes.len() {
         let end_index = min(bytes.len(), line_number * width + width);
-        &bytes[line_number * width..end_index]
+        line_number * width..end_index
     }
-
     else {
-        &[]
+        0..0
     }
 }
 
 
+fn bytes_line(bytes:&[u8], line_number:usize, width:NonZeroUsize) -> &[u8] {
+    &bytes[bytes_line_range(bytes, line_number, width)]
+}
+
+
+fn chars_line(bytes:&[u8], line_number:usize, width:NonZeroUsize) -> String {
+    let mut to_return:String = "".to_owned();
+    for index in bytes_line_range(bytes, line_number, width) {
+        to_return += &String::from(chared_byte(bytes[index], true));
+    }
+    to_return
+}
+
+
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -535,6 +563,14 @@ mod tests {
         assert_eq!(max_bytes_line(&bytes, _4), 0);
     }
 
+    #[test]
+    fn test_num_graphemes() {
+        assert_eq!(num_graphemes("hey, there"), 10);
+        assert_eq!(num_graphemes("दीपक"), 3);
+        assert_eq!(num_graphemes("ﷺ"), 1);
+        assert_eq!(num_graphemes("père"), 4);
+    }
+
     #[test]
     fn test_bytes_line() {
         let bytes = vec![];
@@ -606,6 +642,9 @@ fn print_state(state:&State) {
         println!("Printing byte numbers in {}",
             string_from_radix(state.radix));
     };
+    if state.show_chars {
+        println!("Printing char representations after bytes");
+    };
     println!("Interpreting input numbers as {}",
             string_from_radix(state.radix));
     println!("Printing a newline every {} bytes",
@@ -654,15 +693,37 @@ fn print_bytes(state:&State, range:(usize, usize)) -> Option<usize> {
                 print!("{:>5x}{}", left_col_byte_num, state.n_padding);
             }
         }
-        let cur_line = bytes_line(bytes, bytes_line_num, state.width)
-                .iter().map(|x| formatted_byte(*x, true)).collect::<Vec<String>>().join(" ");
-        println!("{}", cur_line);
+        let cur_line = bytes_line(bytes, bytes_line_num, state.width);
+        let with_color = cur_line.iter().map(|x| formatted_byte(*x, true))
+                .collect::<Vec<String>>().join(" ");
+        let sans_color = cur_line.iter().map(|x| formatted_byte(*x, false))
+                .collect::<Vec<String>>().join(" ");
+        print!("|{}", with_color);
+        let expected_length = usize::from(state.width) * 3 - 1;
+        for _ in num_graphemes(&sans_color)..expected_length {
+            print!(" ");
+        }
+        print!("|");
+        // TODO Do this padding stuff format!  Unclear why previous attempts
+        // have failed.
+        if state.show_chars {
+            print!("   {}", chars_line(bytes, bytes_line_num, state.width));
+        }
+        println!();
         left_col_byte_num = from + bytes_line_num * usize::from(state.width);
     }
     Some(left_col_byte_num)
 }
 
 
+/// .len gives the number of bytes
+/// .chars.count() gives the number of characters (which counts è as two characters.
+/// The human concept is unicode "graphemes" or "glyphs" defined to be what
+/// think they are.
+fn num_graphemes(unicode_string: &str) -> usize {
+    return unicode_string.graphemes(true).count();
+}
+
 fn number_dot_dollar(index:usize, _max_index:usize, input:&str, radix:u32)
         -> Result<usize, String> {
     match input {
@@ -710,6 +771,7 @@ fn lino(state:&State) -> String {
 struct State {
     radix: u32,
     show_byte_numbers: bool,
+    show_chars: bool,
     unsaved_changes: bool,
     filename: String,
 
@@ -728,9 +790,10 @@ struct State {
 
 impl fmt::Debug for State {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "radix: {}|unsaved_changes: {}|show_byte_numbers: {}|index: {}|width: {}|n_padding: '{}'|filename: {}|",
+        write!(f, "radix: {}|unsaved_changes: {}|show_byte_numbers: {}|show_chars: {}|index: {}|width: {}|n_padding: '{}'|filename: {}|",
                 self.radix, self.unsaved_changes, self.show_byte_numbers,
-                self.index, self.width, self.n_padding, self.filename)
+                self.show_chars, self.index, self.width, self.n_padding,
+                self.filename)
     }
 }
 
@@ -804,6 +867,7 @@ pub fn actual_runtime(filename: &str) -> i32 {
         radix: 16,
         filename: filename.to_owned(),
         show_byte_numbers: true,
+        show_chars: true,
         unsaved_changes: false,
         index: 0,
         width: NonZeroUsize::new(16).unwrap(),
@@ -938,6 +1002,12 @@ pub fn actual_runtime(filename: &str) -> i32 {
                         print_bytes(&state, state.range());
                     },
 
+                    /* Toggle showing char representations of bytes */
+                    'm' => {
+                        state.show_chars = !state.show_chars;
+                        println!("{}", state.show_chars);
+                    },
+
                     /* Toggle showing byte number */
                     'n' => {
                         state.show_byte_numbers = !state.show_byte_numbers;
